###################################################################################################
# zmsobject.py
#
# Implementation of class ZMSObject (see below).
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###################################################################################################

# Imports.
from Globals import HTMLFile, Persistent
from DateTime.DateTime import DateTime
from Acquisition import Implicit
import AccessControl.Role
import urllib
import time
import string 
# Product Imports.
import _accessmanager
import _cachemanager
import _copysupport
import _exportable 
import _globals
import _language
import _linklib
import _metapatternmanager
import _objattrs
import _objinputs
import _objtypes
import _versionmanager
import _xmllib
import _zcatalogmanager


###################################################################################################
###################################################################################################
###
###   C l a s s
###
###   This is the abstract base class for all ZMS-Objects.
###
###################################################################################################
###################################################################################################
class ZMSObject(
	Persistent,                     	# Persistent. 
	Implicit,                       	# Acquisition. 
	AccessControl.Role.RoleManager, 	# Security manager. 
	_accessmanager.AccessableObject,	# Access manager.
	_language.Language,             	# Multilinguality.
	_versionmanager.VersionManager,		# Version manager.
	_copysupport.CopySupport,		# Copy Support (Paste Objects).
	_cachemanager.CacheableObject,		# Cacheable object
	_metapatternmanager.MetapatternObject,	# Metapattern object
	_exportable.Exportable,			# XML Export.
	_objattrs.ObjAttrs,			# Object-Attributes.
	_objinputs.ObjInputs,			# Object-Inputs.
	_objtypes.ObjTypes,			# Object-Types
	_zcatalogmanager.ZCatalogItem		# ZCatalog Item
	): 

    # Documentation string.
    __doc__ = """ZMS product module."""
    # Version string. 
    __version__ = '0.1' 
    
    # Management Permissions.
    # -----------------------
    __viewPermissions__ = (
		'manage_menu',
		)
    __ac_permissions__=(
		('View', __viewPermissions__),
		)

    # Properties.
    # -----------
    MISC_ZMS = "misc_/zms/"
    kalender_gif = "misc_/zms/kalender.gif"
    banner_gif = "misc_/zms/banner.gif"
    spacer_gif = "misc_/zms/spacer.gif"
    mail_link_gif = "misc_/zms/mail_link.gif"
    external_link_gif = "misc_/zms/external_link.gif"
    internal_link_gif = "misc_/zms/internal_link.gif"
    internal_link_broken_gif = "misc_/zms/internal_link_broken.gif"

    # Templates.
    # ----------
    manage = HTMLFile('dtml/object/manage', globals())
    manage_workspace = HTMLFile('dtml/object/manage', globals())
    manage_menu = HTMLFile('dtml/object/manage_menu', globals())
    manage_tabs = HTMLFile('dtml/object/manage_tabs', globals()) # Manage Tabulators
    manage_bodyTop = HTMLFile('dtml/object/manage_bodytop', globals()) # Manage bodyTop: Seitenkopf (Logo und berschrift)
    manage_page_header = HTMLFile('dtml/object/manage_page_header', globals()) 
    manage_page_footer = HTMLFile('dtml/object/manage_page_footer', globals()) # Manage bodyBottom: Seitenfuss (Copyright)
    manage_UndoForm = HTMLFile('dtml/object/manage_undoform', globals())
    f_headline = HTMLFile('dtml/object/f_headline', globals()) # berschrift
    f_breadcrumbs = HTMLFile('dtml/object/f_breadcrumbs', globals()) # Manage Breadcrumbs
    f_css_sys = HTMLFile('dtml/object/f_css_sys', globals()) # CSS: Fixed System StyleSheet
    f_css_printhtml = HTMLFile('dtml/object/f_css_printhtml', globals()) # CSS: Fixed PrintHTML StyleSheet
    f_submitBtn = HTMLFile('dtml/object/f_submitbtn', globals()) #
    f_collectionBtn = HTMLFile('dtml/object/f_collectionbtn', globals()) #
    f_languages = HTMLFile('dtml/object/f_languages', globals()) # Content Language select
    f_kalender = HTMLFile('dtml/object/f_kalender', globals()) 
    f_frame = HTMLFile('dtml/object/f_frame', globals()) 
    f_frame_top = HTMLFile('dtml/object/f_frame_top', globals()) 
    f_frame_bottom = HTMLFile('dtml/object/f_frame_bottom', globals()) 
    f_bo_area = HTMLFile('dtml/object/f_bo_area', globals()) 
    f_eo_area = HTMLFile('dtml/object/f_eo_area', globals()) 
    preview_html = HTMLFile('dtml/object/preview', globals()) 
    preview_top_html = HTMLFile('dtml/object/preview_top', globals()) 

    # JavaScript.
    # -----------
    comlib_js = HTMLFile('dtml/javascript/comlib', globals()) # Common Functions
    formlib_js = HTMLFile('dtml/javascript/formlib', globals()) # FORM-Functions
    datelib_js = HTMLFile('dtml/javascript/datelib', globals()) # Date-Functions
    fieldlib_js = HTMLFile('dtml/javascript/fieldlib', globals()) # INPUT-Functions
    managelib_js = HTMLFile('dtml/javascript/managelib', globals()) # Manage-Functions

    # linklib.
    # --------
    browse_objs = HTMLFile('dtml/linklib/browse_objs', globals()) 


    ###############################################################################################
    # ZMSObject.__init__: 
    #
    # Constructor (initialise a new instance of ZMSObject).
    ###############################################################################################
    def __init__(self, id='', sort_id=0): 
      """ ZMSObject.__init__ """
      self.id = id
      self.sort_id = _globals.format_sort_id(sort_id)

    # -------------------------------------------------------------------------------------------------
    #	ZMSObject.getSelf:
    # -------------------------------------------------------------------------------------------------
    def getSelf(self, meta_type=None):
      parent = self.getParentNode()
      if meta_type is not None:
        if type(meta_type) is type([]):
          if self.meta_type not in meta_type and \
             parent is not None:
            return parent.getSelf(meta_type)
        elif type(meta_type) is type(self.PAGES) or \
             type(meta_type) is type(self.PAGEELEMENTS):
          if ((meta_type == self.PAGES and not self.isPage()) or
              (meta_type == self.PAGEELEMENTS and not self.isPageElement())) and \
             parent is not None:
            return parent.getSelf(meta_type)
        else:
          if self.meta_type != meta_type and \
             parent is not None:
            return parent.getSelf(meta_type)
      return self

    # -------------------------------------------------------------------------------------------------
    #	ZMSObject.isPage:
    # -------------------------------------------------------------------------------------------------
    def isPage(self): return 0

    # -------------------------------------------------------------------------------------------------
    #	ZMSObject.isPageElement:
    # -------------------------------------------------------------------------------------------------
    def isPageElement(self): return 0

    # -------------------------------------------------------------------------------------------------
    #	ZMSObject.isActive:
    # -------------------------------------------------------------------------------------------------
    def isActive(self, REQUEST):
      #-- [ReqBuff]: Fetch buffered value from Http-Request.
      try:
        reqBuffId = 'isActive'
        return self.fetchReqBuff(reqBuffId,REQUEST)
      except:
      
        #-- Get value.
        value = 1
        obj_attrs = self.getObjAttrs()
        if value and 'active' in obj_attrs.keys():
          value = value and self.getObjProperty('active',REQUEST)
        # Current time.
        t0 = time.localtime(time.time())
        # Start time.
        if value and 'attr_active_start' in obj_attrs.keys():
          t1 = self.getObjProperty('attr_active_start',REQUEST)
          if t1 is not None:
            value = value and _globals.compareDate(t1,t0) >= 0
        # End time.  
        if value and 'attr_active_end' in obj_attrs.keys():
          t1 = self.getObjProperty('attr_active_end',REQUEST)
          if t1 is not None:
            value = value and _globals.compareDate(t1,t0) < 1
                
        #-- [ReqBuff]: Returns value and stores it in buffer of Http-Request.
        return self.storeReqBuff(reqBuffId,value,REQUEST)

    # -------------------------------------------------------------------------------------------------
    #	ZMSObject.isResource
    # -------------------------------------------------------------------------------------------------
    def isResource(self, REQUEST):
      #-- [ReqBuff]: Fetch buffered value from Http-Request.
      try:
        reqBuffId = 'isResource'
        return self.fetchReqBuff(reqBuffId,REQUEST)
      except:
    
        #-- Get value.
        value = 0
        obj_attrs = self.getObjAttrs()
        if 'attr_dc_type' in obj_attrs.keys():
          attr_dc_type = self.getObjProperty('attr_dc_type',REQUEST)
          value = value or attr_dc_type == 'Resource'
          value = value or attr_dc_type == ['Resource']
    
        #-- [ReqBuff]: Returns value and stores it in buffer of Http-Request.
        return self.storeReqBuff(reqBuffId,value,REQUEST)

    # -------------------------------------------------------------------------------------------------
    #	ZMSObject.isTranslated
    #
    #  Returns 1 if current object is translated to given language.
    # -------------------------------------------------------------------------------------------------
    def isTranslated(self, lang, REQUEST):
      req = {}
      req['lang'] = lang
      if _globals.isPreviewRequest(REQUEST):
        req['preview'] = REQUEST['preview']
      translated = 1
      translated = translated and self.isVisible(req)
      translated = translated and self.getObjProperty('change_uid',req)
      return translated

    # ---------------------------------------------------------------------------------------------
    #  ZMSObject.isVisible:
    #
    #  Returns 1 if current object is visible.
    # ---------------------------------------------------------------------------------------------
    def isVisible(self, REQUEST):
      visible = 1
      visible = visible and self.isActive(REQUEST) # Object is active.
      visible = visible and self.getObjProperty('change_uid',REQUEST) # Object is translated.
      visible = visible and self.isCommitted(REQUEST) # Object has been committed.
      visible = visible and self.getDCCoverage(REQUEST) in ['obligation','proposal',REQUEST['lang']]
      return visible

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.display_icon:
    # ---------------------------------------------------------------------------------------------
    def display_icon(self, REQUEST, meta_type=''):
      if len(meta_type) == 0:
        if self.isActive(REQUEST):
          return self.icon
        else:
          return self.icon_disabled
      else:
        return self.dGlobalAttrs[meta_type]['icon']

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.display_type:
    # ---------------------------------------------------------------------------------------------
    def display_type(self, REQUEST=None, meta_type=''):
      if len(meta_type) == 0:
        meta_type = self.meta_type
      return self.getLangStr("TYPE_%s"%string.upper(meta_type),REQUEST['manage_lang'])

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.breadcrumbs_obj_path:
    # ---------------------------------------------------------------------------------------------
    def breadcrumbs_obj_path(self):
      rtn = []
      obj = self
      for lvl in range(0,self.getLevel()+1):
        obj_item = [obj]
        obj_item.extend(rtn)
        rtn = obj_item
        obj = obj.aq_parent
      return rtn

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.absolute_obj_path:
    # ---------------------------------------------------------------------------------------------
    def absolute_obj_path(self):
      ob = self.getDocumentElement()
      return '%s/%s/'%(ob.aq_parent.id,self.absolute_url()[len(ob.aq_parent.absolute_url())+1:])

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.relative_obj_path:
    # ---------------------------------------------------------------------------------------------
    def relative_obj_path(self):
      return self.absolute_url()[len(self.getDocumentElement().absolute_url())+1:]

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.raw_html_file:
    #
    #	Self-documentation. Key can be any HTMLFile defined for ZMS (e.g. bodyContent_Teaser)
    # ---------------------------------------------------------------------------------------------
    def raw_html_file(self, key, REQUEST, RESPONSE):
      """ raw_html_file """
      filename = getattr(self,key).raw
      f = open(filename,'r')
      html_file = f.read()
      f.close()
      html_file = _globals.html_quote(html_file)
      html_file = string.replace(html_file,'\n','<br>')
      html_file = string.replace(html_file,' ','&nbsp;')
      html_file = string.replace(html_file,'&lt;!--','<font style="color:green">&lt;!--')
      html_file = string.replace(html_file,'--&gt;','--&gt;</font>')
      html = ''
      html = html + '<html>'
      html = html + '<head><title>' + key + '</title></head>'
      html = html + '<body>'
      html = html + '<code>' + html_file + '</code>'
      html = html + '</body>'
      html = html + '</html>'
      return html

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.recurse_afterImport:
    #
    #	Remove helpers after XML-Import.
    # ---------------------------------------------------------------------------------------------
    def _recurse_afterImport(self):
      if hasattr(self,'oParent'): delattr(self,'oParent')
      if hasattr(self,'dTagStack'): delattr(self,'dTagStack')
      if hasattr(self,'dValueStack'): delattr(self,'dValueStack')
    #
    def recurse_afterImport(self):
      self._recurse_afterImport()


    """
    ###############################################################################################
    ###  
    ###  G l o b a l   F u n c t i o n s
    ### 
    ###############################################################################################
    """    
    
    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.rand_int:
    # ---------------------------------------------------------------------------------------------
    def rand_int(self, n):
      from whrandom import randint
      return randint(0,n)

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.url_quote:
    # ---------------------------------------------------------------------------------------------
    def url_quote(self, s):
      return urllib.quote(s)
    
    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.url_inherit_params:
    # ---------------------------------------------------------------------------------------------
    def url_inherit_params(self, s, REQUEST, exclude=[]):
      rtn = s
      if REQUEST.form:
        count = 0
        for key in REQUEST.form.keys():
          if count == 0:
            rtn = rtn + "?"
          else:
            rtn = rtn + "&"
          rtn = rtn + key + "=" + urllib.quote(REQUEST.form[key])
          count = count + 1    	  
      return rtn

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.id_quote:
    # ---------------------------------------------------------------------------------------------
    def id_quote(self, s):
      return _globals.id_quote(s)

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.js_quote:
    # ---------------------------------------------------------------------------------------------
    def js_quote(self, s):
      return _globals.js_quote(s)

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.isPreviewRequest:
    # ---------------------------------------------------------------------------------------------
    def isPreviewRequest(self, REQUEST):
      return _globals.isPreviewRequest(REQUEST)


    """
    ###############################################################################################
    ###  
    ###  C o m m o n   F u n c t i o n s
    ### 
    ###############################################################################################
    """    
    
    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.recurse_updateVersion:
    #
    #	Update version.
    # ---------------------------------------------------------------------------------------------
    def recurse_updateVersion(self, lang, manage_lang, REQUEST): 
        message = ''

        # Process object.
        # [...]
        
        # ---------------------------------------------------------------------
        # 18.02.2002 Build #058
        # Multilingual Modification User-ID and Time.
        # ---------------------------------------------------------------------
        attrs = {'uid':'change_uid','modification_time':'change_dt'}
        for s_attr in attrs.keys():
          if hasattr(self,s_attr):
            for s_preview in [None,'preview']:
              req = {'preview':s_preview}
              ob = self.getObjVersion(req)
              for s_lang in self.getLanguages():
                #++ print "%s.%s_%s=%s"%(str(ob),attrs[s_attr],s_lang,str(getattr(self,s_attr)))
                setattr(ob,'%s_%s'%(attrs[s_attr],s_lang),getattr(self,s_attr))
            delattr(self,s_attr)

        # Return with message.
        return message


    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.findObjId:
    #
    #	Searches for Object identified by ID and returns the absolute URL of it.
    # ---------------------------------------------------------------------------------------------
    def _findObjId(self, id, REQUEST):
      rtn = None
      if str(self.id) == str(id):
        rtn = self
      elif self.isPage():
        for child in self.getChildNodes(REQUEST):
          if rtn is None:
            rtn = child._findObjId(id,REQUEST)
      return rtn
    # --
    def findObjId(self, relative_obj_path, REQUEST):
      root = self.getDocumentElement()
      rtn = root
      if len(relative_obj_path) > 0:
        # Find object by ids given in object path.
        ids = string.split(relative_obj_path,"/")
        for id in ids:
          if rtn is not None:
            rtn = getattr(rtn,id,None)
        # Find object by id.
        if rtn is None:
          rtn = root._findObjId(ids[-1],REQUEST) 
      # Return object.
      return rtn

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.href2Html:
    #	ZMSObject.getHref2IndexHtml:
    #	ZMSObject.getHref2SitemapHtml:
    #
    #   "Sans-Document"-Navigation: reference to first ZMSContainerObject
    #	that contains visible page-elements.
    # ---------------------------------------------------------------------------------------------
    def getHref2Html(self, fct, REQUEST):
      root = self.getDocumentElement().aq_parent
      href = self.absolute_url()
      href = href + '/%s_%s.html'%(fct,REQUEST['lang'])
      if _globals.isPreviewRequest(REQUEST): href = href + '?preview=preview'
      return href
    #++
    def findHref2Html(self, fct, REQUEST):
      if not self.isPage():
        parent = self.getParentNode()
        return '%s#%s'%(parent.getHref2Html(fct,REQUEST),self.id)
      else:
        cntnr = None
        for ob in self.getChildNodes(REQUEST):
          if ob.isVisible(REQUEST): 
            if ob.isPageElement():
              return self.getHref2Html(fct,REQUEST)
            elif ob.isPage():
              if cntnr is None: cntnr = ob
        if cntnr is not None:
          return cntnr.findHref2Html(fct,REQUEST)
      return self.getHref2Html(fct,REQUEST)
    #++
    def getHref2SitemapHtml(self, REQUEST): 
      if not REQUEST.has_key('lang'): REQUEST.set('lang',self.getPrimaryLanguage())
      return self.getHref2Html('sitemap',REQUEST)
    #++
    def getHref2PrintHtml(self, REQUEST): 
      if not REQUEST.has_key('lang'): REQUEST.set('lang',self.getPrimaryLanguage())
      qs = REQUEST.get('QUERY_STRING','')
      if len(qs) > 0:
        qs = '?' + qs
      return self.getHref2Html('index_print',REQUEST) + qs
    #++
    def getHref2IndexHtml(self, REQUEST): 

      #-- [ReqBuff]: Fetch buffered value from Http-Request.
      try:
        reqBuffId = 'getHref2IndexHtml'
        value = self.fetchReqBuff(reqBuffId,REQUEST)
        return value
      except:
      
        #-- Get value.
        if not REQUEST.has_key('lang'): REQUEST.set('lang',self.getPrimaryLanguage())
        value = self.findHref2Html('index',REQUEST)
      
        #-- [ReqBuff]: Returns value and stores it in buffer of Http-Request.
        return self.storeReqBuff(reqBuffId,value,REQUEST)
        

    """
    ###############################################################################################    
    ###  
    ###  Auswahlliste fr Aktionen 
    ### 
    ###############################################################################################    
    """

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.filtered_edit_actions:
    #
    #	Object-actions of management interface.
    # ---------------------------------------------------------------------------------------------
    def filtered_edit_actions(self, path, REQUEST):
      rtn = []
      
      lang = REQUEST['lang']
      manage_lang = REQUEST['manage_lang']
      auth_user = REQUEST['AUTHENTICATED_USER']
      
      if self.getAutocommit() or \
         self.getPrimaryLanguage() == lang or \
         self.getDCCoverage(REQUEST) not in ['obligation','proposal']:
        actions = []
        actions.append((self.getLangStr('BTN_DELETE',manage_lang),'manage_deleteObjs',[]))
        actions.append((self.getLangStr('BTN_CUT',manage_lang),'manage_cutObjects',[]))
        actions.append((self.getLangStr('BTN_COPY',manage_lang),'manage_copyObjects',[]))
        if self.cb_dataValid():
          actions.append((self.getLangStr('BTN_PASTE',manage_lang),'manage_pasteObjs',[]))
        actions.append((self.getLangStr('ACTION_MOVEUP',manage_lang),path + 'manage_moveObjUp',[]))
        actions.append((self.getLangStr('ACTION_MOVEDOWN',manage_lang),path + 'manage_moveObjDown',[]))
        if len(actions) > 0:
          rtn.append((self.getLangStr('ACTION0',manage_lang),'',actions))
          
      # Return action list.
      return rtn

    
    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.filtered_child_actions:
    #
    #	Object-actions of management interface.
    # ---------------------------------------------------------------------------------------------
    def filtered_child_actions(self, REQUEST):
      rtn = []
      
      lang = REQUEST['lang']
      manage_lang = REQUEST['manage_lang']
      auth_user = REQUEST['AUTHENTICATED_USER']
      
      if self.meta_type != 'ZMSSysFolder' or \
         auth_user.has_permission('ZMS Administrator',self.getParentNode()):
        rtn.extend(self.getParentNode().filtered_edit_actions(self.id + '/',REQUEST))
        rtn.extend(self.getParentNode().filtered_insert_actions(self.id + '/',REQUEST))
        
      # Return action list.
      return rtn


    """
    ###############################################################################################
    ###  
    ###  Check In / Out
    ### 
    ###############################################################################################
    """

    checkout_overtake_html = HTMLFile('dtml/object/checkout_overtake', globals()) 
    checkout = HTMLFile('dtml/object/checkout', globals()) 

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.checkIn:
    # ---------------------------------------------------------------------------------------------
    def checkIn(self, REQUEST):
      if hasattr(self,'co_dat_%s'%REQUEST['lang']): delattr(self,'co_dat_%s'%REQUEST['lang'])
      if hasattr(self,'co_uid_%s'%REQUEST['lang']): delattr(self,'co_uid_%s'%REQUEST['lang'])

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.checkOut:
    # ---------------------------------------------------------------------------------------------
    def checkOut(self, REQUEST):
      setattr(self,'co_dat_%s'%REQUEST['lang'],time.time())
      setattr(self,'co_uid_%s'%REQUEST['lang'],str(REQUEST['AUTHENTICATED_USER']))

    """
    ###############################################################################################
    ###  
    ###  Sitemap
    ### 
    ###############################################################################################
    """

    ###############################################################################################
    #  ZMSContainerObject.manage_sitemap: 
    #
    #  De-/Activate sitemap.
    ###############################################################################################
    def manage_sitemap(self, lang, manage_lang, REQUEST, RESPONSE):
      """ ZMSObject.manage_sitemap """
      sitemap = 1
      if REQUEST.has_key('dtpref_sitemap'):
        sitemap = not string.atoi(REQUEST['dtpref_sitemap'])
      e=(DateTime('GMT') + 365).rfc822()
      RESPONSE.setCookie('dtpref_sitemap',str(sitemap),path='/',expires=e)
      return RESPONSE.redirect('manage?lang=%s&manage_lang=%s'%(lang,manage_lang))


    """
    ###############################################################################################
    ###  
    ###  Related Links 
    ### 
    ###############################################################################################
    """

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.getLinkObj:
    #
    #	Resolves internal/external links and returns Object.
    # ---------------------------------------------------------------------------------------------
    def getLinkObj(self, url, REQUEST): 
      return _linklib.getLinkObj(self,url,REQUEST)

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.getLinkUrl:
    #
    #	Resolves internal/external links and returns URL.
    # ---------------------------------------------------------------------------------------------
    def getLinkUrl(self, url, REQUEST): 
      return _linklib.getLinkUrl(self,url,REQUEST)


    """
    ###############################################################################################    
    ###  
    ###  Catalog (Suche)
    ### 
    ###############################################################################################    
    """

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.getCatalogNavUrl:
    #
    #	Returns catalog-navigation url.
    # ---------------------------------------------------------------------------------------------
    def getCatalogNavUrl(self,REQUEST):
      rtn = REQUEST['URL']
      rtn = '%s?op=%s'%(rtn,urllib.quote(REQUEST['op']))
      rtn = '%s&raw=%s'%(rtn,urllib.quote(REQUEST['raw']))
      rtn = '%s&option=%s'%(rtn,urllib.quote(REQUEST['option']))
      rtn = '%s&orderby=%s'%(rtn,urllib.quote(str(REQUEST['orderby'])))
      return rtn


    """
    ###############################################################################################    
    ###  
    ###  DOM-Methoden 
    ### 
    ###############################################################################################    
    """    

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.getLevel:
    #
    #	The hierarchical level of this node.
    # ---------------------------------------------------------------------------------------------
    def getLevel(self, level=0):
      parent = self.getParentNode()
      if parent is not None and getattr(parent,'meta_type','') in self.dGlobalAttrs.keys():
        level = parent.getLevel(level+1)
      return level

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.isAnchestor:
    #
    #	True if this object is anchestor of given object.
    # ---------------------------------------------------------------------------------------------
    def isAnchestor(self, ob):
      return string.find(ob.absolute_url() + "/",self.absolute_url() + "/") == 0

    # ---------------------------------------------------------------------------------------------
    #  ZMSObject.getParentByDepth:
    #
    #  The parent of this node by depth. 
    # ---------------------------------------------------------------------------------------------
    def getParentByDepth(self, deep):
        rtn = self
        for i in range(0,deep):
          rtn = rtn.getParentNode()
        return rtn

    # ---------------------------------------------------------------------------------------------
    #  ZMSObject.getParentByLevel:
    #
    #  The parent of this node by level. 
    # ---------------------------------------------------------------------------------------------
    def getParentByLevel(self, level):
        rtn = self
        while rtn.getLevel() > level:
          rtn = rtn.getParentNode()
        return rtn

    # ---------------------------------------------------------------------------------------------
    #  ZMSObject.getParentNode:
    #
    #  The parent of this node. 
    #  All nodes except the root object may have a parent.
    # ---------------------------------------------------------------------------------------------
    def getParentNode(self):
        parent = None
        if hasattr(self,'aq_parent'):
          parent = getattr(self,'aq_parent')
        return parent

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.getChildNodes:
    #
    #	Returns a NodeList that contains all children of this node, 
    #	if none, this is a empty NodeList. 
    # ---------------------------------------------------------------------------------------------
    def getChildNodes(self, REQUEST=None, meta_types=None): 
      return []


    """
    ###############################################################################################    
    ###  
    ###  Sortier-Reihenfolge
    ### 
    ###############################################################################################    
    """

    # ---------------------------------------------------------------------------------------------
    #	ZMSObject.getSortId:
    #
    #	Returns Sort-ID (integer).
    # ---------------------------------------------------------------------------------------------
    def getSortId(self,REQUEST): 
      s = self.getObjProperty('sort_id',REQUEST)
      return string.atoi(s[len(_globals.id_prefix(s)):])


    ###############################################################################################    
    # ZMSObject.manage_moveObjUp:
    #
    # Moves an Object up in sort order.
    ###############################################################################################    
    def manage_moveObjUp(self, lang, manage_lang, REQUEST, RESPONSE): 
      """ ZMSObject.manage_moveObjUp """
      parent = self.getParentNode()
      sort_id = self.getSortId(REQUEST)
      self.setObjStateModified(REQUEST)
      self.setObjProperty('sort_id',_globals.format_sort_id(sort_id - 15))
      parent.normalizeSortIds(_globals.id_prefix(self.id))
      # Return with message.
      message = self.getLangStr('MSG_MOVEDOBJUP',manage_lang)%("<i>%s</i>"%self.display_type(REQUEST))
      RESPONSE.redirect('%s/manage_main?lang=%s&manage_lang=%s&manage_tabs_message=%s#_%s'%(parent.absolute_url(),lang,manage_lang,urllib.quote(message),self.id))


    ###############################################################################################    
    # ZMSObject.manage_moveObjDown:
    #
    # Moves an Object down in sort order.
    ###############################################################################################    
    def manage_moveObjDown(self, lang, manage_lang, REQUEST, RESPONSE): 
      """ ZMSObject.manage_moveObjDown """
      parent = self.getParentNode()
      sort_id = self.getSortId(REQUEST)
      self.setObjStateModified(REQUEST)
      self.setObjProperty('sort_id',_globals.format_sort_id(sort_id + 15))
      parent.normalizeSortIds(_globals.id_prefix(self.id))
      # Return with message.
      message = self.getLangStr('MSG_MOVEDOBJDOWN',manage_lang)%("<i>%s</i>"%self.display_type(REQUEST))
      RESPONSE.redirect('%s/manage_main?lang=%s&manage_lang=%s&manage_tabs_message=%s#_%s'%(parent.absolute_url(),lang,manage_lang,urllib.quote(message),self.id))


    """
    ###############################################################################################    
    #
    #   Darstellung
    #
    ###############################################################################################    
    """

    # ---------------------------------------------------------------------------------------------
    #  ZMSObject.getBodyContent:
    #         
    #  HTML presentation in BodyContent. 
    # ---------------------------------------------------------------------------------------------
    def _getBodyContent(self, REQUEST, thumbnail=0): return ''
    
    def getBodyContent(self, REQUEST):
      rtn = ""
      lang = REQUEST['lang']
      if self.isVisible(REQUEST):
        rtn = rtn + '<a name="%s"></a>'%self.id
        rtn = rtn + self._getBodyContent(REQUEST)
      return rtn


    # ---------------------------------------------------------------------------------------------
    # ZMSObject.printHtml:
    #
    # Renders print presentation of a ContainerObject.
    # ---------------------------------------------------------------------------------------------
    def printHtml(self, level, sectionize, sectionizer, REQUEST):
      return self._getBodyContent(REQUEST)


    """
    ###############################################################################################    
    #
    #   BUILDER
    #
    ###############################################################################################    
    """
    
    ###############################################################################################
    # ZMSObject.xmlOnStartElement(self, dTagName, dAttrs, oCurrNodes, oRoot):
    # ZMSObject.xmlOnCharacterData(self, data, bInCData):
    # ZMSObject.xmlOnEndElement(self):
    # ZMSObject.xmlOnUnknownStartTag(self, sTagName, dTagAttrs)
    # ZMSObject.xmlOnUnknownEndTag(self, sTagName)
    # ZMSObject.xmlGetTagName(self):
    # ZMSObject.xmlGetParent(self):
    #
    # handler for XML-Builder (_builder.py)
    ###############################################################################################
    def xmlOnStartElement(self, sTagName, dTagAttrs, oParentNode, oRoot):
        _globals.myDebugger.write("[%s.xmlOnStartElement]: sTagName=%s"%(self.meta_type,sTagName))
        
        self.dTagStack    = _globals.MyStack()
        self.dValueStack  = _globals.MyStack()
        
	#-- ATTRIBUTES
	for i in range(0,len(dTagAttrs)/2):
	  key = dTagAttrs[i*2]
	  if key != 'id':
  	    value = dTagAttrs[i*2+1]
	    #++ print "[%s.xmlOnStartElement]: %s=%s"%(self.meta_type,str(key),str(value))
	    setattr(self,key,value)
        
        # WORKAROUND! The member variable "aq_parent" does not contain the right parent object at 
        # this stage of the creation process (it will later on!). Therefore, we introduce a special
        # attribute containing the parent object, which will be used by xmlGetParent() (see below).
        self.oParent = oParentNode


    def xmlOnEndElement(self): 
        pass


    def xmlOnCharacterData(self, sData, bInCData):
        _xmllib.xmlOnCharacterData(self,sData,bInCData)
        pass


    def xmlOnUnknownStartTag(self, sTagName, dTagAttrs):
        return _xmllib.xmlOnUnknownStartTag(self,sTagName,dTagAttrs)


    def xmlOnUnknownEndTag(self, sTagName):
        return _xmllib.xmlOnUnknownEndTag(self,sTagName)


    def xmlGetParent(self):
        return self.oParent


    def xmlGetTagName(self):
        return self.meta_type

###################################################################################################